
#include "qobjects.h"

#include <cstdlib>
#include <ctime>
		  
//---------------------------------------initialize parameters

const int seed = 234562;               // random seed  23456

const double hbar = 1,                 // fundamental constants
			 e  = 2.7182818,
			 pi = 3.1415927; 
	
const std::complex<double> ei(0,1); 

const unsigned int Nfock = 15,         // number of fock states 
				   ns = 10000,         // number of time steps 
				   ns_noise = ns;    
const int          state_out = ns, 
				   series_out = ns/2000; 

const double T = 4,                 // total time 
			 dt = T/ns,        
	  
			 w  = 2*pi,     // omega 
			 g  = 4,        // damping rate 
			 nt = 1,        // average thermal phonons
			 beta = std::log(1 + 1/nt)/w;       // stat mech beta 
		   
qoperator I("identity",Nfock), 
		  a("destroy",Nfock), 
		  adag("create",Nfock), 
		  N = adag*a, 
		  X3 = N*N, 
		  x = std::sqrt(0.5)*(a + adag), 
		  x2 = x*x, 
		  p = ei*std::sqrt(0.5)*(adag - a),
		  p2 = p*p; 
		  
//-------------------------------- chose a Hamiltonian

qoperator H = w*X3,    
		  H_ho = w*N,  
		  dU = (-ei*dt)*H; 
   
//-------------------------------- Master Equations 

// quantum optics ME 
qmatrix QOME( qmatrix& rho) 
{ 
    qmatrix sym_rhoE = N*rho + rho*N; 
	qmatrix drho = dU*rho - rho*dU 
				 + ((nt+1)*(g/4)*dt)*( (2*a)*rho*adag - sym_rhoE ) 
			     + ((nt)*(g/4)*dt)*( 2*(adag*rho*a - rho) - sym_rhoE ); 
	return drho; 
}

// Low Temperature Brownian motion ME 
qmatrix LBME( qmatrix& rho) 
{ 
    qmatrix sym_rhop = p*rho + rho*p;
	qmatrix drho = dU*rho - rho*dU 
	             - ((g/4)*(2*nt+1)*dt)*( x2*rho + rho*x2 - 2*x*rho*x )  
				 - (ei*(g/4)*dt)*( x*sym_rhop - sym_rhop*x ); 
	return drho; 
}

// Positive (Lindblad) Brownian motion ME 
qmatrix PosBME( qmatrix& rho) 
{ 
    qmatrix sym_rhop = p*rho + rho*p;
	qmatrix drho = dU*rho - rho*dU 
	             - ((g/4)*(2*nt+1)*dt)*( x2*rho + rho*x2 - 2*x*rho*x )  
				 - (g/(16*(2*nt+1))*dt)*( p2*rho + rho*p2 - 2*p*rho*p )
				 - (ei*(g/4)*dt)*( x*sym_rhop - sym_rhop*x ); 
	return drho; 
}

//--------------------------------- choose a Master Equation 
qmatrix (*ME) (qmatrix&) = &LBME; 


int main(int argc, char* argv[]){ 
   
//--------------------------------initialise files

	std::ofstream  densr("rhor.dat",std::ios_base::trunc);
	std::ofstream  densi("rhoi.dat",std::ios_base::trunc);
	std::ofstream  thermr("rhor_therm.dat",std::ios_base::trunc);
	std::ofstream  thermi("rhoi_therm.dat",std::ios_base::trunc);	
	std::ofstream  thermr_ho("rhor_therm_ho.dat",std::ios_base::trunc);
	std::ofstream  thermi_ho("rhoi_therm_ho.dat",std::ios_base::trunc);	
	std::ofstream  series("series.dat",std::ios_base::trunc);
	std::ofstream  avs("avs.dat",std::ios_base::trunc);

//--------------------------------initial state 

	qstate psi("coherent",0,Nfock);
	qmatrix rho(psi);
	   
    // output initial state 
	rho.output_1line(densr,densi);
   
	// output initial series point 
	double meanx, Vx, meanp, Vp; 
	meanx = trace(x*rho).real(); 
	Vx = trace(x2*rho).real() - meanx*meanx; 
	meanp = trace(p*rho).real(); 
	Vp = trace(p2*rho).real() - meanp*meanp; 
	series << 0 << " " << meanx << " " << meanp << " " 
					   << Vx    << " " << Vp    << " " << trace(N*rho).real() << std::endl; 

//--------------------------------thermal state (for comparison) 

	qoperator therm = H; 
			  therm.exp_diag(-beta); 
	qmatrix rho_therm(therm); 
		    rho_therm.normalise();  

	// output thermal state 
	rho_therm.output_1line(thermr,thermi);
	
//-------------------------------- H.O. thermal state (for comparison) 

	qoperator therm_ho = H_ho; 
			  therm_ho.exp_diag(-beta); 
	qmatrix rho_therm_ho(therm_ho); 
		    rho_therm_ho.normalise();  

	// output H.O. thermal state 
	rho_therm_ho.output_1line(thermr_ho,thermi_ho);  
	
//-------------------------------- output average n for thermal states 
	double n_therm = trace(N*rho_therm).real(); 
	double n_therm_ho = trace(N*rho_therm_ho).real(); 
	avs << n_therm << " " << n_therm_ho << " " << std::endl; 
   
//------------------------------------------------------------
//                        Simulation
//------------------------------------------------------------
          
	time_t time0, *time_ptr = 0;       // timing initialisation 
	time0 = time(time_ptr);
            
	// integrate a single trajectory
	for (unsigned int i=0;i<ns;i++) {            
	  
		// report progress
		if (std::div(i,500).rem==0) {std::cout << "T = " << i*dt << std::endl;}  

		RK4_stepper(ME,rho);  
		rho.normalise();     
		
		// output snapshots 
        if (std::div(i+1,state_out).rem==0) { rho.output_1line(densr,densi); }
		
		// output time series 
		if (std::div(i+1,series_out).rem==0) { 
			meanx = trace(x*rho).real(); 
			Vx = trace(x2*rho).real() - meanx*meanx; 
			meanp = trace(p*rho).real(); 
			Vp = trace(p2*rho).real() - meanp*meanp; 
			series << (i+1)*dt << " " << meanx << " " << meanp 
							   << " " << Vx    << " " << Vp << " " << trace(N*rho).real() << std::endl; 
		}
		
	}

   //-----------------------
   // report simulation time
   //-----------------------

   std::cout << std::endl << "Simulation took " << difftime(time(time_ptr),time0) 
                          << " seconds" << std::endl;
                    
   // tidy up 
   densr.close(); thermr.close(); thermr_ho.close(); 
   densr.close(); thermi.close(); thermi_ho.close();
   series.close(); avs.close(); 
   	                 
   return 0;
}
