
#include "qobjects.h"
#include "mpi.h" 
#include <ctime> 

//---------------------------------------------------------------------------------------
int    global_state( std::vector<qstate> &psi, qmatrix& rho, int my_rank, int nproc); 
int    thisproc_state( std::vector<qstate> &psi, qmatrix& rho, int nproc); 
double global_mean( const std::vector<qstate> &psi_vec, qoperator &Op, int nproc); 
double thisproc_mean( const std::vector<qstate> &psi_vec, qoperator &Op, int nproc); 
int    output_state( std::vector<qstate>& psi, std::ofstream& densr, 
					 std::ofstream& densi, int my_rank, int nproc); 

int thermal(qstate& psi, fftw_plan& plan1,      fftw_plan& plan2, 
						 std::vector<double>& p, std::vector<double>& x, 
						 double hbar, double dt, double w, double G, double n_T); 
//----------------------------------------------------------------------------------------


int main(int argc, char* argv[]){ 

//-------------------------------- initialize MPI

	int my_rank, nproc;
	MPI_Status status;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
	MPI_Comm_size(MPI_COMM_WORLD,&nproc);
   
//------------------------------initialize parameters

   const int seed = 1256;             

   const double hbar = 1.0,             // fundamental constants
                e  = 2.7182818,
                pi = 3.1415927; 

   const std::complex<double> zi(0,1);  
   
   const double  w = 2*pi; 

   const unsigned int nx = 128,            // points in x-grid
					  HO_N_max = 20,       // number of Fock states in basis 
					  
                      ns_per_Nsteps = 4,   // 
					   
					  which_series = 0;    
					  									  
   const double tstart = 0,          // time window to output the state    
				tfin = 1000;		 // 
				
   const double n_T = 1,	        // average thermal phonons  (1.62=10mK, 5.76=30mK, 4.24=22.7mK)
				av_n_init = 0, 
				beta = std::log(1 + 1/n_T)/w;	    // stat mech beta 
				
				// ---------------------------------------------------- 
   
   const double L = std::sqrt(nx*2*pi),   // need this L to give correct aspect ratio for wignerX.m 
                dx = L/nx, 
                dp = 2*pi*hbar/L, 
                Lp = nx*dp,
   
                meanx0 = 0,               // initial state parameters 
                meanp0 = 0,               
                Vx0 = hbar/2,
                Vp0 = hbar/2, 
				
				Fock_scaling = 1.0,       // Fock-state parameter (should be 1)
				rescale = 1.0/nx;         // FFT normalization 

//-------------------------------- pars/variables for monte-carlo decoherence 

	const int    traj_per_proc = 320,  // 
			     tot_traj = traj_per_proc*nproc; 
				 
//--------------------------------initialize files

	std::ofstream  densr, densi;
	std::string realfile = "rhor.dat";  
	std::string imagfile = "rhoi.dat";  
	densr.open(realfile.c_str(),std::ios_base::trunc);
	densi.open(imagfile.c_str(),std::ios_base::trunc); 	
	
	std::ofstream  thermr("rhor_therm.dat",std::ios_base::trunc);
	std::ofstream  thermi("rhoi_therm.dat",std::ios_base::trunc);	
	std::ofstream  thermr_ho("rhor_therm_ho.dat",std::ios_base::trunc);
	std::ofstream  thermi_ho("rhoi_therm_ho.dat",std::ios_base::trunc);	
	
	std::ofstream  series("series.dat",std::ios_base::trunc);		
	
	std::ofstream  ss_avs("ss_avs.dat",std::ios_base::trunc);	
	
// ------------------------------------- timing initialisation
				 
		time_t time0, *time_ptr = 0; 
		time0 = time(time_ptr);

//--------------------------------initialize grids and functions  

	std::vector<double> x,p,Fx2,Fp2;
	std::vector< std::complex<double> > FMx(nx);
										 
	for (unsigned int i=0;i<nx;++i) {     // x and functions of x 
		x.push_back(-L/2 + i*dx);
		Fx2.push_back(x[i]*x[i]);
	}
					// p and functions of p in fftshifted space
	for (unsigned int i=nx/2;i<nx;++i) { 
		p.push_back(-Lp/2 + i*dp);
	}
	for (unsigned int i=0;i<nx/2;++i) { 
		p.push_back(-Lp/2 + i*dp);
	}
	for (unsigned int i=0;i<nx;++i) { 
		Fp2.push_back(p[i]*p[i]);
	}
   
//--------------------------------initialize basic operators

    qoperator Ident("identity",nx),  
              Xx(x),                   Xx2(Fx2),  	   
              Pp(p),                   Pp2(Fp2); 
			 			 			 			 	
	qoperator To_HO("transform:x->Fock",x,Fock_scaling,HO_N_max); 
	qoperator From_HO = Hconj(To_HO); 
		
//-------------------------------------- thermal state (for comparison)

	qoperator a("destroy",HO_N_max), adag("create",HO_N_max), 
			  N = adag*a, 
			  X3 = N*N, 
			  x_op = std::sqrt(0.5)*(a + adag), 
			  x2_op = x_op*x_op;  

	qoperator therm = w*X3; 
			  therm.exp_diag(-beta); 
	qmatrix rho_therm(therm); 
			rho_therm.normalise();  
			
	double meanx_therm = trace(x_op*rho_therm).real(); 
	double Vx_therm = trace(x2_op*rho_therm).real() - meanx_therm*meanx_therm; 

 
	rho_therm.output_1line(thermr,thermi);
		
//-------------------------------------- H.O. thermal state (for comparison)	

	qoperator therm_ho = w*N; 
			  therm_ho.exp_diag(-beta); 
	qmatrix rho_therm_ho(therm_ho); 
			rho_therm_ho.normalise();  
			
	rho_therm_ho.output_1line(thermr_ho,thermi_ho); 
				
//--------------------------------------------------------
//               Loop over parameter values 
//-------------------------------------------------------- 

	int parnum = 9;   // number of parameter values to loop over 
	
	std::vector<double> G_set(parnum); 
	double G_min = 7.9433*4, G_max = 39.8107*4, G_del = std::exp(std::log(G_max/G_min)/(parnum-1)); 
	G_set[0] = G_min; 
	for (int par=1;par<parnum;par++) { G_set[par] = G_set[par-1]*G_del; } 
	
	std::vector<double> T_set(parnum); 
    double T_max = 1.5, T_min = 0.5, T_del = std::exp(std::log(T_min/T_max)/(parnum-1)); 
	T_set[0] = T_max; 
	for (int par=1;par<parnum;par++) { T_set[par] = T_set[par-1]*T_del; } 
	
	std::vector<double> NsPUT_set(parnum);    
    double NsPUT_min = 2000, NsPUT_max = 8000, NsPUT_del = std::exp(std::log(NsPUT_max/NsPUT_min)/(parnum-1)); 
	NsPUT_set[0] = NsPUT_min; 
	for (int par=1;par<parnum;par++) { NsPUT_set[par] = NsPUT_set[par-1]*NsPUT_del; } 

	//------------------------------------------------ 
	//----------------- begin loop -------------------
	
	for (int par=0;par<parnum;par++) {  

		double G = G_set[par];  
		double T = T_set[par]; 
		unsigned int Nsteps = static_cast<unsigned int>(std::ceil(T*NsPUT_set[par])); 
		
		if (my_rank==0) { std::cout << Nsteps << " " << T << std::endl; } 
							  
	    unsigned int series_out = 60,    // output controls 
		             state_out = Nsteps;  // this choice outputs the final state only 
					  									  
		double Ndt = T/Nsteps, 
				dt = Ndt/ns_per_Nsteps;  
				
		// diagonal evolution operators 
      	qoperator UX = 0.5*w*Xx2; UX.exp_it_diag(-dt);
		qoperator UP = 0.5*w*Pp2; UP.exp_it_diag(-dt);
		
		std::vector< std::complex<double> >  dU_HO_vec(HO_N_max);  
		std::vector< std::complex<double> >  dU_X3_vec(HO_N_max);  
		for (int j=0;j<HO_N_max;j++) { 
			dU_HO_vec[j] = std::cos(j*w*Ndt/hbar) - zi*std::sin(j*w*Ndt/hbar); 
			dU_X3_vec[j] = std::cos(j*j*w*Ndt/hbar) - zi*std::sin(j*j*w*Ndt/hbar);
		} 
		qoperator dU_HO(dU_HO_vec); 
		qoperator dU_X3(dU_X3_vec); 
		
		//------------------------------- choose the Hamiltonian

		qoperator dU = dU_X3;     // dimension nx 

		//--------------------------------initial state (thermal) 

		std::srand(seed + 100*my_rank);
		std::vector<qstate> psi(traj_per_proc); 
		
		for (int traj=0;traj<traj_per_proc;traj++) {
			std::vector<double> randnum(2);
			make_noise(randnum);
			double xtherm = randnum[0]*std::sqrt(hbar*av_n_init);
			double ptherm = randnum[1]*std::sqrt(hbar*av_n_init);
			qstate psi0("Gaussian",nx,L,hbar,Vx0,Vp0,meanx0 + xtherm, meanp0 + ptherm);
			psi[traj] = psi0; 
		}		
		
		qmatrix rho_sim(nx);
				
		// ------------------------------------------- fftw initialization

		fftw_plan plan1[traj_per_proc], plan2[traj_per_proc];
		for (int traj=0;traj<traj_per_proc;traj++) { psi[traj].fftw_init(plan1[traj], plan2[traj]); }
							
		//-------------------------------------- output the initial state (commented out)
		
		std::vector<qstate> psi_HO(traj_per_proc);
//		for (int traj=0;traj<traj_per_proc;traj++) { psi_HO[traj] = To_HO*psi[traj]; } 
//		output_state(psi_HO,densr,densi,my_rank,nproc); 
//		//output_state(psi,densr,densi,my_rank,nproc); 

		//-------------------------------------- output the first time series point 
	
		double meanx, Vx, meanp, meanp2, Vp; 
		meanx = global_mean(psi,Xx,nproc); 
		Vx = global_mean(psi,Xx2,nproc) - meanx*meanx;
		if (my_rank==0) { 
			series << 0 << " " << meanx << " " << Vx << " " << 0 << " " << 0.5 << " " << av_n_init << std::endl; 
		} 
				
		//-----------------------------------------------------------------------
		//                              Simulation
		//-----------------------------------------------------------------------

		for (int Ni=0;Ni<Nsteps;Ni++) { 
		
			double t = Ni*Ndt; 
			
			// Hamiltonian (this is exact evolution)  
			for (int traj=0;traj<traj_per_proc;traj++) { 
				psi[traj].apply(To_HO); psi[traj].apply(dU);   
				psi[traj].apply(From_HO); psi[traj].normalise(); 
			} 

			for (int i=0;i<ns_per_Nsteps;i++) { 
			   
				t = Ni*Ndt + i*dt; 
										
				for (int traj=0;traj<traj_per_proc;traj++) { 
					// Hamiltonian 
	//				psi[traj].apply(UX); 
	//				psi[traj].fftw(plan1[traj]); psi[traj].apply(UP); psi[traj].fftw(plan2[traj]); 
	//				psi[traj].normalise(); 
		
					// Thermal noise 
					thermal(psi[traj], plan1[traj], plan2[traj], p, x, hbar, dt, w, G, n_T);
					psi[traj].normalise();
								
				}
				
			} 
						
			//------------------------------- output time series 
			MPI_Barrier(MPI_COMM_WORLD);
			if (std::div(Ni+1,series_out).rem==0) {
			
				meanx = global_mean(psi,Xx,nproc); 
				double meanx2 = global_mean(psi,Xx2,nproc);
				Vx = meanx2 - meanx*meanx; 
				
				for (int traj=0;traj<traj_per_proc;traj++) { psi[traj].fftw(plan1[traj]); }
				meanp = rescale*global_mean(psi,Pp,nproc);
				meanp2 = rescale*global_mean(psi,Pp2,nproc);
				for (int traj=0;traj<traj_per_proc;traj++) { psi[traj].fftw(plan2[traj]); psi[traj].normalise(); }
				
				Vp = meanp2 - meanp*meanp; 
				
				double meanN = (meanx2 + meanp2)/2 - 0.5; 
				if (my_rank==0) { series << t + dt << " " << meanx << " " << Vx << " " 
				                                          << meanp << " " << Vp << " " << meanN << std::endl; } 
			} 

			//------------------------------------- output snapshots 

			if (std::div(Ni+1,state_out).rem==0 && t >= tstart && t <= tfin) { 
				for (int traj=0;traj<traj_per_proc;traj++) { psi_HO[traj] = To_HO*psi[traj]; } 
				output_state(psi_HO,densr,densi,my_rank,nproc);
				//output_state(psi,densr,densi,my_rank,nproc);
			}

			
		} 
		//__________________________________________________________________________
		//______________________________End Simulation______________________________  

		//-------------------------------- output ss averages 		
		double mx2 = global_mean(psi,Xx2,nproc);
		double mx2_1 = thisproc_mean(psi,Xx2,nproc); 
		
		for (int traj=0;traj<traj_per_proc;traj++) { psi[traj].fftw(plan1[traj]); }
		double mp2 = rescale*global_mean(psi,Pp2,nproc); 
		double mp2_1 = rescale*thisproc_mean(psi,Pp2,nproc); 
		for (int traj=0;traj<traj_per_proc;traj++) { psi[traj].fftw(plan2[traj]); psi[traj].normalise(); }
		
		double meanN_1 = (mx2_1 + mp2_1)/2 - 0.5; 
		double meanN   = (mx2   + mp2  )/2 - 0.5;
				
		double N_err_1 = (meanN_1 - meanN)*(meanN_1 - meanN); 
		double N_err; 
		MPI_Allreduce(&N_err_1,&N_err,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
		N_err = std::sqrt(N_err/(nproc*nproc)); 
		
		qmatrix rhof, rhof_1; 
		for (int traj=0;traj<traj_per_proc;traj++) { psi_HO[traj] = To_HO*psi[traj]; } 
		thisproc_state(psi_HO,rhof_1,nproc); rhof_1.truncate(HO_N_max);
		
		double meanN2_1 = trace(X3*rhof_1).real(); 
		double meanN2;  
		MPI_Allreduce(&meanN2_1,&meanN2,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
		meanN2 /= nproc; 

		double N2_err_1 = (meanN2_1 - meanN2)*(meanN2_1 - meanN2); 
		double N2_err; 
		MPI_Allreduce(&N2_err_1,&N2_err,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD); 
		N2_err = std::sqrt(N2_err/(nproc*nproc)); 
							
		if (my_rank==0) { 
			ss_avs << G_set[par] << " " << 
		//  FBME             therm nonlin                         therm lin 
			meanN  << " " <<
			N_err  << " " << trace(N*rho_therm).real()  << " " << trace(N*rho_therm_ho).real()  << " " << 
			meanN2 << " " << 
			N2_err << " " << trace(X3*rho_therm).real() << " " << trace(X3*rho_therm_ho).real() << std::endl; 
		}
							 
		// tidy up fftw
		for (int traj=0;traj<traj_per_proc;traj++) { psi[traj].fftw_fin(plan1[traj],plan2[traj]); }

	} 
	//__________________________________________________________________________
	//______________________End Loop Over Parameter Values______________________ 
	
	
	// report simulation time  
	MPI_Barrier(MPI_COMM_WORLD);
	double hours = difftime(time(time_ptr),time0)/3600.0;
	double minutes = hours*60; 
	if (my_rank==0) { 
		std::cout << std::endl << "  Time taken: " << hours << " hours" << std::endl; 
		std::cout << std::endl << "          or  " << minutes << " mins" << std::endl; 
	} 
   
	// tidy up
	densr.close(); 
	densi.close(); 
	MPI_Finalize(); 

	return 0;
}


// ----------------------- subroutines -------------------------------

int global_state( std::vector<qstate> &psi, qmatrix& rho, int my_rank, int nproc) 
{
	MPI_Status status;
	int traj_per_proc = psi.size(); 
	std::vector<unsigned int> dims = psi[0].get_dims(); 
	int nx = dims[0]; 
	qstate psi_rec = psi[0]; // any state with the right dimension will do here 
	qstate rho_new(nx); 
	rho = rho_new;

	for (int traj=0;traj<traj_per_proc;traj++) { 
		for (int np=1;np<nproc;np++) {  		   
			MPI_Barrier(MPI_COMM_WORLD);
			if (my_rank == np) { MPI_Send(psi[traj].get_array(),2*nx,MPI_DOUBLE,0,99,MPI_COMM_WORLD); } 
			if (my_rank == 0) { 
				MPI_Recv(psi_rec.get_array(),2*nx,MPI_DOUBLE,np,99,MPI_COMM_WORLD,&status); 
				qmatrix rho_rec(psi_rec); 
				rho += rho_rec; 
			} 
		} 
	} 
	
	if (my_rank==0) { 
		for (int traj=0;traj<traj_per_proc;traj++) {
		   qmatrix rho_temp(psi[traj]); 
		   rho += rho_temp; 
		}
	}
	
	rho.normalise(); 	
	return 0; 
}

int thisproc_state( std::vector<qstate> &psi, qmatrix& rho, int nproc)
{
	int traj_per_proc = psi.size(); 
	std::vector<unsigned int> dims = psi[0].get_dims(); 
	int nx = dims[0]; 
	qstate rho_new(nx); 
	rho = rho_new;

	for (int traj=0;traj<traj_per_proc;traj++) { 
		qmatrix rho_next(psi[traj]); 
		rho += rho_next;
	} 
	
	rho.normalise(); 	
	return 0; 
}

double global_mean( const std::vector<qstate> &psi_vec, qoperator &Op, int nproc)
{
	double sub_meanM(0), meanM(0); 
	int traj_per_proc = psi_vec.size(); 
	double norm = 1.0/(nproc*traj_per_proc); 
	for (int traj=0;traj<traj_per_proc;traj++) { 
		sub_meanM += Op.expect(psi_vec[traj]).real(); 
	}
	MPI_Allreduce(&sub_meanM,&meanM,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
	return norm*meanM; 
}

double thisproc_mean( const std::vector<qstate> &psi_vec, qoperator &Op, int nproc)
{
	double meanM(0); 
	int traj_per_proc = psi_vec.size(); 
	for (int traj=0;traj<traj_per_proc;traj++) { 
		meanM += Op.expect(psi_vec[traj]).real(); 
	}
	return meanM/traj_per_proc; 
}

int output_state( std::vector<qstate>& psi, std::ofstream& densr, std::ofstream& densi, int my_rank, int nproc) 
{
	MPI_Status status;
	int traj_per_proc = psi.size(); 
	std::vector<unsigned int> dims = psi[0].get_dims(); 
	int nx = dims[0]; 
	qstate psi_rec = psi[0]; // any state will do here 
	qmatrix rho(psi_rec);
	rho.clear();

	for (int traj=0;traj<traj_per_proc;traj++) { 
		for (int np=1;np<nproc;np++) {  		   
			MPI_Barrier(MPI_COMM_WORLD);
			if (my_rank == np) {MPI_Send(psi[traj].get_array(),2*nx,MPI_DOUBLE,0,99,MPI_COMM_WORLD); } 
			if (my_rank == 0) { 
				MPI_Recv(psi_rec.get_array(),2*nx,MPI_DOUBLE,np,99,MPI_COMM_WORLD,&status); 
				qmatrix rho_rec(psi_rec); 
				rho += rho_rec; 
			} 
		} 
	} 
	
	rho.normalise(); 
	if (my_rank == 0) { rho.output_1line(densr,densi); } 
	
	return 0; 
}

int thermal(qstate& psi, fftw_plan& plan1,      fftw_plan& plan2, 
						 std::vector<double>& p, std::vector<double>& x, 
						 double hbar, double dt, double w, double gamma, double n_T) 
{
	static bool firstcall = true; 
	static const std::complex<double> zi(0,1); 
	static double meanx, Vx, meanp, Vp, Cxp; 
	double G = gamma*n_T; 
	static std::vector<double> noiseT(2); 
	
	static qstate psi_o(psi), psi_x2(psi), psi_p(psi), psi_xp(psi), psi_p2(psi); 
	static fftw_plan plan_xp1, plan_xp2; 
	if (firstcall) { psi_xp.fftw_init(plan_xp1,plan_xp2); }

	
	static qoperator Xx(x), Pp(p), Xx2 = Xx*Xx, Pp2 = Pp*Pp; 
	static unsigned int nx = x.size();
	static double rescale = 1.0/nx; 
	static 	std::vector< std::complex<double> > Fx2(nx), Fp2(nx), Gx(nx), Gp(nx), Ux(nx), Up(nx); 
	if (firstcall) { for (int j=0;j<nx;j++) { Fx2[j] = x[j]*x[j]; Fp2[j] = p[j]*p[j]; } }
	
	make_noise(noiseT); 
	
	// ====== Step 1:
	meanx = (Xx.expect(psi)).real(); 
	Vx = (Xx2.expect(psi)).real() - meanx*meanx;

	psi_o = psi;
	psi_x2 = psi;
	psi_x2.apply(Xx2);
	
	// ====== Step 2: Transform Psi etc 
	psi.fftw(plan1);

	meanp = rescale*(Pp.expect(psi)).real();
	Vp = rescale*(Pp2.expect(psi)).real() - meanp*meanp; 

	psi_p = psi;
	psi_p.apply(Pp);
	psi_p2 = psi_p;
	psi_p2.apply(Pp); 
	psi_xp = psi_p;
	psi_xp.fftw(plan_xp2); psi_xp = rescale*psi_xp; 
	psi_xp.apply(Xx); 
	
	Cxp = (psi_o.inner_prod(psi_xp)).real() - meanx*meanp; 
					
	// ====== Step 3: 
	double rGdW1 = std::sqrt(G*dt)*noiseT[0];
	double rGdW2 = std::sqrt(G*dt)*noiseT[1];

	for (unsigned int j=0;j<nx;++j) {             
		Gp[j] = std::exp(  (-zi*dt/hbar)*0.5*w*Fp2[j] + 
						   - G*dt*Fp2[j] + 2*G*dt*p[j]*meanp + p[j]*rGdW2 );
		Up[j] = std::exp( 2.0*zi*(Vx*rGdW1 + Cxp*rGdW2)*p[j] );
	}
	qoperator Gp_op(Gp);
	qoperator Up_op(Up);
	psi.apply(Gp_op);
	psi.apply(Up_op);
	
	// ====== Step 4: 
	psi.fftw(plan2); psi = rescale*psi;
	
	for (unsigned int j=0;j<nx;++j) {              
		Gx[j] = std::exp( ( -zi*dt/hbar)*( 0.5*w*Fx2[j] + (0.5*G/n_T)*meanp*x[j] ) 
						   - G*dt*Fx2[j] + 2*G*dt*x[j]*meanx + x[j]*rGdW1 );
		Ux[j] = std::exp( -2.0*zi*(Vp*rGdW2 + (Cxp-0.5)*rGdW1)*x[j] );
	}
	qoperator Gx_op(Gx);
	qoperator Ux_op(Ux);
	psi.apply(Gx_op);
	psi.apply(Ux_op);   
	
	if (firstcall) { firstcall = false; }
	return 0;
}

